@extends('layouts.meeting')
@section('content')

    <div id="proCallComponent">

        <div class="d-flex align-items-center justify-content-center py-3 border-bottom">
            <span class="mr-3">
                {{ $pro->name_display }} | PRO
            </span>
            <button class="btn btn-sm btn-primary px-4 font-weight-bold"
                    v-on:click.prevent="nextPatient()"
                    :disabled="client || checkingForNextPatient || started">Next Patient</button>
            <span v-if="patientInQueue && !started" class="patient-in-q-alert text-warning text-sm ml-2 small">
                <i class="fa fa-circle"></i>
            </span>
            <span v-if="!patientInQueue && !started" class="text-success text-sm ml-2 small">
                <i class="fa fa-circle"></i>
            </span>
        </div>
        <div v-if="!started && noNextPatient" class="bg-light rounded text-center py-1 font-weight-bold text-sm my-3 mx-3">@{{ noNextPatient }}</div>

        <div class="" v-show="videoActive">
            <div class="py-3 text-center" v-if="started">
                <h6 class="text-black font-weight-bold m-0">Call in progress: @{{ timeDisplay() }}</h6>
            </div>
            <div class="py-3 text-center" v-if="noOneElseInCall">
                <h6 class="text-black font-weight-bold m-0">No other participants in the call.
                    <a href="#" class="text-danger font-weight-bold" v-on:click.prevent="hangUp()">Hang up</a>
                </h6>
            </div>
            <div class="main-view mx-auto">
                <div id="self-view" class="full-view" data-name="{{ $pro->name_display }}" data-type="PRO"></div>
                <div class="thumbs">

                </div>
                <button class="btn btn-danger rounded-circle hang-up"
                        v-if="started"
                        title="Leave Call"
                        v-on:click.prevent="hangUp()">
                    <i class="fa fa-phone"></i>
                </button>
            </div>
        </div>

    </div>

    <script>

        new Vue({
            el: '#proCallComponent',
            delimiters: ['@{{', '}}'],
            data: {
                time: 0,
                startTime: 0,
                started: false,
                client: false,
                pro: false,

                selfName: '',
                selfToken: '',

                clientUid: '',

                otSessionId: '',

                checkingForNextPatient: false,
                noNextPatient: false,

                otSession: false,

                selfUserType: 'PRO',
                selfStreamId: '',
                noOneElseInCall: true,
                patientInQueue: false,

                videoActive: false,

                heartbeatTimer: false,
            },
            methods: {
                pollForNextPatient: function() {
                    if(!this.started) {
                        this.nextPatient(true);
                    }
                },
                nextPatient: function(_pollOnly = false) {
                    var self = this;
                    if(!_pollOnly) this.checkingForNextPatient = true;
                    $.post('/api/client/getNextClientForVideoVisit', {}, function(_data) {
                        if(_pollOnly) {
                            self.patientInQueue = _data.success;
                        }
                        else {
                            self.checkingForNextPatient = false;
                            if(!_data.success) {
                                self.noNextPatient = _data.message;
                                window.setTimeout(function() {
                                    self.noNextPatient = false;
                                }, 2000);
                            }
                            else {
                                // get ot session key from client record
                                self.client = true;
                                self.clientUid = _data.data;
                                self.videoActive = true;
                                self.startOpenTokSession();
                            }
                        }
                    }, 'json');
                },
                startOpenTokSession: function() {
                    var self = this;
                    self.getOpenTokSessionId(function() {
                        self.selfName = '{{ $pro->name_display  }}';
                        $.post('/api/openTok/getClientToken', {
                            opentokSessionId: self.otSessionId,
                            data: JSON.stringify({
                                uid: '{{ $pro->uid  }}',
                                name: self.selfName,
                                type: 'PRO'
                            })
                        }, function (_data) {
                            console.log(_data);
                            self.selfToken = _data.data;
                            self.initOpenTok();
                        });
                    });
                },
                timeDisplay: function() {
                    var seconds = this.time / 1000,
                        minutes = parseInt(seconds / 60, 10);
                    seconds = parseInt(seconds % 60, 10);
                    return minutes + " min, " + seconds + " sec";
                },
                hangUp: function() {
                    if(this.otSession) {
                        try {
                            this.otSession.disconnect();
                        }
                        catch (e) {
                            console.log('Was already disconnected.');
                        }
                        this.otSession = false;
                        this.otSessionId = '';
                        this.started = false;
                        this.startTime = false;
                        this.videoActive = false;
                        // this.client = false;

                        // call leaveVideoVisitAsPro
                        $.post('/api/clientVideoVisit/leaveVideoVisitAsPro', {}, function(_data) {
                            console.log(_data);
                        });
                    }
                },
                initOpenTok: function() {

                    /* fake video feed (temp) */
                    const randomColour = () => {
                        return Math.round(Math.random() * 255);
                    };

                    const canvas = document.createElement('canvas');
                    canvas.width = 640;
                    canvas.height = 480;
                    const ctx = canvas.getContext('2d');
                    var pos = 100;
                    window.setInterval(function() {
                        ctx.clearRect(0, 0, canvas.width, canvas.height);
                        ctx.font = "20px Georgia";
                        ctx.fillStyle = `rgb(220, 220, 220)`;
                        ctx.fillText("Video feed from {{ $pro->name_display }}", 20, pos);
                        pos += 5;
                        if(pos > canvas.height) pos = 100;
                    }, 1000);

                    var self = this;

                    var apiKey = '<?=  config('stag.tokboxApiKey') ?>';
                    var sessionId = this.otSessionId;
                    var token = this.selfToken;

                    self.otSession = OT.initSession(apiKey, sessionId);

                    // peer connected
                    self.otSession.on('streamCreated', function streamCreated(event) {
                        console.log('streamCreated', arguments);
                        var subscriberOptions = {
                            insertMode: 'append',
                            width: '100%',
                            height: '100%'
                        };

                        var connectionData = JSON.parse(event.stream.connection.data);

                        // add a div for remove view
                        var remoteViewID = 'remote-view-' + event.stream.id;
                        var remoteElem = $('<div id="' + remoteViewID + '" class="remote-view thumb-view" ' +
                            'data-stream="' + event.stream.id + '" ' +
                            'data-connection-data="' + event.stream.connection.data + '" ' +
                            'data-name="' + connectionData.name + '" ' +
                            'data-type="' + connectionData.type + '"></div>');
                        remoteElem.appendTo('.thumbs');

                        self.otSession.subscribe(event.stream, remoteViewID, subscriberOptions, self.handleOpenTokError);

                        if (connectionData.type === 'CLIENT') {
                            self.client = true;
                        }

                        if(!self.startTime) {
                            self.startTime = new Date().getTime();
                            window.setInterval(function() {
                                self.time = new Date().getTime() - self.startTime;
                            }, 1000);
                            self.started = true;
                        }

                        self.activateParty(event.stream.id);

                        self.noOneElseInCall = false;
                    });

                    // peer disconnected
                    self.otSession.on("streamDestroyed", function(event) {
                        onPeerDisconnection(event, event.stream.connection.data);
                    });
                    // self.otSession.on("connectionDestroyed", function(event) {
                    //     debugger;
                    //     console.log('connectionDestroyed from ' + event.connection.data);
                    //     onPeerDisconnection(event, event.connection.data);
                    // });

                    self.otSession.on("connectionCreated", function(event) {
                        console.log('connectionCreated');
                        console.log(event);
                    });

                    function onPeerDisconnection(event, data) {

                        if(event.stream && $('.full-view[data-stream="' + event.stream.id + '"]').length) {
                            var allThumbs = $('.thumbs [data-stream]:not([data-stream=""]):not(.disconnected-view):visible');
                            if(allThumbs.length) {
                                $('.thumbs [data-stream]:not([data-stream=""])').each(function() {
                                    if($(this).attr('data-stream') !== event.stream.id) {
                                        self.activateParty($(this).attr('data-stream'));
                                        return false;
                                    }
                                });
                            }
                            else {
                                self.noOneElseInCall = true;
                            }
                        }

                        if(event.stream) {
                            var remoteViewElem = $('[data-stream="' + event.stream.id + '"]');
                            remoteViewElem.remove();
                            // if(remoteViewElem.length) {
                            //     remoteViewElem.attr('data-stream', '');
                            //     remoteViewElem.attr('data-connection-data', '');
                            //     remoteViewElem.attr('data-type', '');
                            //     remoteViewElem.attr('data-name', '');
                            // }
                            // remoteViewElem.addClass('disconnected-view')
                        }

                        var connectionData = JSON.parse(data);
                        if(connectionData.type === 'CLIENT') {
                            self.client = false;
                        }

                        // if no other parties in call, hang up
                        if(!$('[data-stream]:not([data-stream="' + self.selfStreamId + '"])').length) {
                            self.startTime = 0;
                            self.started = false;
                            self.noOneElseInCall = true;
                        }
                    }

                    // self connected
                    self.otSession.on("sessionConnected", function(event) {
                        self.joinMeetingAsPro(self.selfUserType);
                    });

                    // self disconnected
                    self.otSession.on('sessionDisconnected', function sessionDisconnected(event) {
                        console.log('You were disconnected from the session.', event.reason);

                        // turn pro video off
                        $.post('/api/clientVideoVisit/turnProVideoOff', {}, function(_data) {
                            console.log(_data);

                            // stop heart beat
                            if(self.heartbeatTimer) {
                                window.clearInterval(self.heartbeatTimer);
                                self.heartbeatTimer = false;
                            }
                        });
                    });

                    // initialize the publisher
                    var publisherOptions = {
                        // videoSource: canvas.captureStream(1).getVideoTracks()[0], // TODO: Comment this line to use webcam
                        insertMode: 'append',
                        width: '100%',
                        height: '100%',
                    };
                    var publisher = OT.initPublisher('self-view', publisherOptions, self.handleOpenTokError);

                    publisher.on('streamCreated', function(event) {
                        var selfView = $('#self-view');
                        selfView.attr('data-stream', event.stream.id);
                        selfView.attr('data-connection-data', event.stream.connection.data);
                        self.selfStreamId = event.stream.id;
                    });

                    publisher.on('streamCreated', function(event) {
                        console.log('publisher->streamCreated');
                        var selfView = $('#self-view');
                        selfView.attr('data-stream', event.stream.id);
                        selfView.attr('data-connection-data', event.stream.connection.data);
                        selfView.attr('data-type', 'PRO');
                        self.activateParty('self');
                        $('#self-view').show();

                        // turn pro video on
                        $.post('/api/clientVideoVisit/turnProVideoOn', {}, function(_data) {
                            console.log(_data);

                            // start heart beat
                            self.heartbeatTimer = window.setInterval(function() {
                                $.post('/api/clientVideoVisit/registerProMeetingHeartbeat', {}, function(_data) {
                                    console.log(_data);
                                });
                            }, 5000);
                        });
                    });

                    publisher.on('streamDestroyed', function(event) {
                        event.preventDefault();
                        console.log('publisher->streamDestroyed');
                        $('#self-view').hide();
                        var allThumbs = $('.thumbs [data-stream]:not([data-stream=""]):not(.disconnected-view):visible');
                        if(allThumbs.length) {
                            $('.thumbs [data-stream]:not([data-stream=""])').each(function() {
                                if($(this).attr('data-stream') !== $('#self-view').attr('data-stream')) {
                                    self.activateParty($(this).attr('data-stream'));
                                    return false;
                                }
                            });
                        }
                        else {
                            self.startTime = 0;
                            self.started = false;
                            self.noOneElseInCall = true;
                        }

                        // turn pro video off
                        $.post('/api/clientVideoVisit/turnProVideoOff', {}, function(_data) {
                            console.log(_data);

                            // stop heart beat
                            if(self.heartbeatTimer) {
                                window.clearInterval(self.heartbeatTimer);
                                self.heartbeatTimer = false;
                            }
                        });
                    });


                    // Connect to the session
                    self.otSession.connect(token, function callback(error) {
                        if (error) {
                            self.handleOpenTokError(error);
                        } else {
                            // If the connection is successful, publish the publisher to the session
                            self.otSession.publish(publisher, self.handleOpenTokError);
                        }
                    });
                },
                handleOpenTokError: function(e) {

                },

                getOpenTokSessionId: function(_done) {
                    var self = this;
                    $.get('/pro/get-opentok-session-key/' + self.clientUid, function(_data) {
                        self.otSessionId = _data.data;
                        console.log(_data);
                        _done();
                    }, 'json');
                },

                joinMeetingAsPro: function(_type) {
                    var self = this;
                    $.ajax({
                        type: 'post',
                        url: '/api/clientVideoVisit/joinVideoVisitAsPro',
                        headers: {
                            'sessionKey': '{{ request()->cookie('sessionKey') }}'
                        },
                        data: {uid: self.clientUid},
                        dataType: 'json'
                    })
                        .done(function (_data) {
                            console.log(_data);
                        })
                        .fail(function (_data) {
                            console.warn(_data);
                            alert(_data.message);
                        });
                },

                activateParty: function(_stream = 'self') {
                    var current = $('.full-view');
                    if(current.attr('data-stream') === _stream) return;
                    current.removeClass('full-view').addClass('thumb-view');
                    if(current.attr('data-type') === 'CLIENT') {
                        current.prependTo('.thumbs');
                    }
                    else {
                        current.appendTo('.thumbs');
                    }
                    if(_stream === 'self') {
                        $('#self-view')
                            .removeClass('thumb-view')
                            .removeClass('disconnected-view')
                            .addClass('full-view')
                            .prependTo('.main-view');
                    }
                    else {
                        $('div[data-stream="' + _stream + '"]')
                            .removeClass('thumb-view')
                            .removeClass('disconnected-view')
                            .addClass('full-view')
                            .prependTo('.main-view');
                    }
                }

            },
            mounted: function() {

                var self = this;

                $(document).on('click', '.thumbs>div[data-stream]', function() {
                    self.activateParty($(this).attr('data-stream'));
                    return false;
                });

                // poll for new patients and alert
                window.setInterval(function() {
                    self.pollForNextPatient();
                }, 5000);

                window.onbeforeunload = function() {
                    if(self.started) {
                        return "A call is in progress";
                    }
                };

                @if(isset($client))
                    self.client = true;
                    self.clientUid = '{{ $client->uid }}';
                    self.videoActive = true;
                    self.startOpenTokSession();
                @endif
            }
        });
    </script>

@endsection
